# The MIT License (MIT)
#
# Copyright (c) 2015 Stefan Fischer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This is a library for reading ARPA n-gram models.

The package may be imported directly::

    import arpa

Details about the ARPA n-gram format can be found here:

- `SRILM <http://www.speech.sri.com/projects/srilm/manpages/ngram-format.5.html>`_
- `ICSI Speech <http://www1.icsi.berkeley.edu/Speech/faq/grammarfmts.html>`_

The library was initiated by Stefan Fischer and is developed and maintained by many others.
"""

from .arpa import dump, dumpf, dumps, load, loadf, loads

__all__ = ["dump dumpf dumps load loadf loads".split()]

__author__ = "Stefan Fischer"
__contact__ = "Stefan Fischer <sfischer13@ymail.com>"
__copyright__ = "Copyright (c) 2015 Stefan Fischer"
__credits__ = []
__date__ = "2015-07-24"
__license__ = "MIT"
__status__ = "development"
__version__ = "0.1.0a6"



from collections import OrderedDict

from .base import ARPAModel


class ARPAModelSimple(ARPAModel):
    def __init__(self):
        self._counts = OrderedDict()
        self._ps = OrderedDict()
        self._bos = OrderedDict()

    def add_count(self, order, count):
        self._counts[order] = count

    def add_entry(self, ngram, p, bo=None, order=None):
        key = tuple(ngram)
        self._ps[key] = p
        if bo:
            self._bos[key] = bo

    def counts(self):
        return sorted(self._counts.items())

    def order(self):
        return max(self._counts.keys())

    def vocabulary(self):
        return sorted(set(word for ngram in self._ps.keys() for word in ngram))

    def _entries(self, order):
        return (self._entry(k) for k in self._ps.keys() if len(k) == order)

    def _entry(self, ngram):
        if ngram in self._bos:
            return self._ps[ngram], ngram, self._bos[ngram]
        else:
            return self._ps[ngram], ngram

    def _log10_bo(self, ngram):
        return self._bos[ngram]

    def _log10_p(self, ngram):
        return self._ps[ngram]



from io import StringIO

from .models.simple import ARPAModelSimple
from .parsers.quick import ARPAParserQuick


def dump(obj, fp):
    """Serialize obj to fp (a file-like object) in ARPA format."""
    obj.write(fp)


def dumpf(obj, path, mode="wt", encoding=None):
    """Serialize obj to path in ARPA format."""
    with open(path, mode=mode, encoding=encoding) as f:
        dump(obj, f)


def dumps(obj):
    """Serialize obj to an ARPA formatted str."""
    with StringIO() as f:
        dump(obj, f)
        return f.getvalue()


def load(fp, model=None, parser=None):
    """Deserialize fp (a file-like object) to a Python object."""
    if not model:
        model = "simple"
    if not parser:
        parser = "quick"

    if model not in ["simple"]:
        raise ValueError
    if parser not in ["quick"]:
        raise ValueError

    if model == "simple" and parser == "quick":
        return ARPAParserQuick(ARPAModelSimple).parse(fp)
    else:
        raise ValueError


def loadf(path, mode="rt", encoding=None, model=None, parser=None):
    """Deserialize path (a text file) to a Python object."""
    with open(path, mode=mode, encoding=encoding) as f:
        return load(f, model, parser)


def loads(s, model=None, parser=None):
    """Deserialize s (a str) to a Python object."""
    with StringIO(s) as f:
        return load(f, model, parser)



from abc import ABCMeta, abstractmethod


class ARPAParser(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, model):
        pass

    @abstractmethod
    def parse(self, fp):
        pass



try:
    from enum import Enum, unique
except ImportError:
    from enum34 import Enum, unique
import re

from ..exceptions import ParseException
from .base import ARPAParser


class ARPAParserQuick(ARPAParser):
    @unique
    class State(Enum):
        DATA = 1
        COUNT = 2
        HEADER = 3
        ENTRY = 4

    re_count = re.compile(r"^ngram (\d+)=(\d+)$")
    re_header = re.compile(r"^\\(\d+)-grams:$")
    re_entry = re.compile("^(-?\\d+(\\.\\d+)?([eE]-?\\d+)?)"
                          "\t"
                          "(\\S+( \\S+)*)"
                          "(\t(-?\\d+(\\.\\d+)?)([eE]-?\\d+)?)?$")

    def __init__(self, model):
        self.ModelClass = model

    def parse(self, fp):
        self._result = []
        self._state = self.State.DATA
        self._tmp_model = None
        self._tmp_order = None
        for line in fp:
            line = line.strip()
            if self._state == self.State.DATA:
                self._data(line)
            elif self._state == self.State.COUNT:
                self._count(line)
            elif self._state == self.State.HEADER:
                self._header(line)
            elif self._state == self.State.ENTRY:
                self._entry(line)
        if self._state != self.State.DATA:
            raise ParseException(line)
        return self._result

    def _data(self, line):
        if line == "\\data\\":
            self._state = self.State.COUNT
            self._tmp_model = self.ModelClass()
        else:
            pass  # skip comment line

    def _count(self, line):
        match = self.re_count.match(line)
        if match:
            order = match.group(1)
            count = match.group(2)
            self._tmp_model.add_count(int(order), int(count))
        elif not line:
            self._state = self.State.HEADER  # there are no counts
        else:
            raise ParseException(line)

    def _header(self, line):
        match = self.re_header.match(line)
        if match:
            self._state = self.State.ENTRY
            self._tmp_order = match.group(1)
        elif line == "\\end\\":
            self._result.append(self._tmp_model)
            self._state = self.State.DATA
            self._tmp_model = None
            self._tmp_order = None
        elif not line:
            pass  # skip empty line
        else:
            raise ParseException(line)

    def _entry(self, line):
        match = self.re_entry.match(line)
        if match:
            p = float(match.group(1))
            ngram = tuple(match.group(4).split(" "))
            bo = match.group(7)
            bo = float(bo) if bo else None
            self._tmp_model.add_entry(ngram, p, bo, self._tmp_order)
        elif not line:
            self._state = self.State.HEADER  # last entry
        else:
            raise ParseException(line)



from abc import ABCMeta, abstractmethod


class ARPAModel(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        pass

    def __contains__(self, word):
        return word in self.vocabulary()

    @abstractmethod
    def add_count(self, order, count):
        pass

    @abstractmethod
    def add_entry(self, ngram, p, bo=None, order=None):
        pass

    def log10_p(self, ngram):
        ngram = self._check_input(ngram)
        try:
            return self._log10_p(ngram)
        except KeyError:
            try:
                log10_bo = self._log10_bo(ngram[:-1])
            except KeyError:
                log10_bo = 0
            return log10_bo + self.log10_p(ngram[1:])

    def log10_s(self, sentence):
        words = self._check_input(sentence)
        return sum(self.log10_p(words[:i]) for i in range(1, len(words) + 1))

    def prob(self, ngram):
        return 10 ** self.log10_p(ngram)

    p = prob

    def score(self, sentence):
        return 10 ** self.log10_s(sentence)

    s = score

    @abstractmethod
    def counts(self):
        pass

    @abstractmethod
    def order(self):
        pass

    @abstractmethod
    def vocabulary(self):
        pass

    # TODO: write comment in header
    def write(self, fp):
        fp.write("\\data\\\n")
        for order, count in self.counts():
            fp.write("ngram {}={}\n".format(order, count))
        fp.write("\n")
        for order, _ in self.counts():
            fp.write("\\{}-grams:\n".format(order))
            for e in self._entries(order):
                fp.write("{}\t{}".format(e[0], " ".join(e[1])))
                if len(e) == 3:
                    fp.write("\t{}".format(e[2]))
                fp.write("\n")
            fp.write("\n")
        fp.write("\\end\\\n")

    @abstractmethod
    def _entries(self, order):
        pass

    @abstractmethod
    def _log10_bo(self, ngram):
        pass

    @abstractmethod
    def _log10_p(self, ngram):
        pass

    @staticmethod
    def _check_input(input):
        if not input:
            raise ValueError
        elif isinstance(input, tuple):
            return input
        elif isinstance(input, list):
            return tuple(input)
        elif isinstance(input, str):
            return tuple(input.strip().split(" "))
        else:
            raise ValueError



"""Exceptions raised by this package."""


class ARPAException(Exception):
    """Common base class for all package exceptions."""
    pass


class ParseException(ARPAException):
    """ARPA file could not be parsed."""
    pass
